/*
  Name: md2.h
  Author: Carsten Haubold (CarstenHaubold@web.de)
  Date: 14.02.05 20:46
  Description: a lot of material to use MD2 Models, I collected it from books abd webpages and changed it so it fits into my engine
*/


#ifndef __MD2_H
#define __MD2_H

#include <windows.h>			// standard Windows app include
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gl/gl.h>				// standard OpenGL include
#include <gl/glu.h>				// OpenGL utilties
#include "texture.h"
#include "3dmath.h"

#define MAX_FRAMES 512





void CalculateNormal( float *p1, float *p2, float *p3 );


/* 
	MD2 Model Functionality
*/
typedef struct{
        char name[50];//str for md2 file
        char tex[50];//str for texture
        char info[50];//str for file containing informations
} modelfile_t;

// texture coordinate
typedef struct
{
   float s;
   float t;
} texCoord_t;

// texture coordinate index
typedef struct
{
   short s;
   short t;
} stIndex_t;

// info for a single frame point
typedef struct
{
   unsigned char v[3];
   unsigned char normalIndex;	// not used
} framePoint_t;

// information for a single frame
typedef struct
{
   float scale[3];
   float translate[3];
   char name[16];
   framePoint_t fp[1];
} frame_t;

// data for a single triangle
typedef struct
{
   unsigned short meshIndex[3];		// vertex indices
   unsigned short stIndex[3];		// texture coordinate indices
} mesh_t;

typedef struct
{
   int ident;		 // identifies as MD2 file "IDP2"
   int version;	 // mine is 8
   int skinwidth;    // width of texture
   int skinheight;   // height of texture
   int framesize;    // number of bytes per frame
   int numSkins;     // number of textures
   int numXYZ;       // number of points
   int numST;        // number of texture
   int numTris;      // number of triangles
   int numGLcmds;
   int numFrames;    // total number of frames
   int offsetSkins;  // offset to skin names (64 bytes each)
   int offsetST;     // offset of texture s-t values
   int offsetTris;   // offset of triangle mesh
   int offsetFrames; // offset of frame data (points)
   int offsetGLcmds; // type of OpenGL commands to use
   int offsetEnd;    // end of file
} modelHeader_t;


class CMD2Model
{
private:

     
     int numVertices;         // number of vertices
     int numTriangles;        // number of triangles
     int numST;               // number of skins
     int frameSize;           // size of each frame in bytes
     int currentFrame;        // current frame # in animation
     int nextFrame;           // next frame # in animation
     float interpol;          // percent through current frame
     mesh_t *triIndex;        // triangle list
     texCoord_t *st;          // texture coordinate list
     Vertex *vertexList;    // vertex list
     texture_t modelTex;     // texture data
    
     GLuint modellist;
     bool alpha;
     



public:
       char name[50];
      int numFrames;			// number of model frames
      



void init()
{
     numVertices = 0;    // vertices
     numTriangles = 0;   // triangles
     numFrames = 0;      // frames
     numST = 0;          // texture coordinates
     frameSize = 0;      // needed?
     currentFrame = 0;   // current keyframe 
     nextFrame = 1;      // next keyframe
     interpol = 0.0;     // interpolation percent
     triIndex = NULL;    // triangle indices
     st = NULL;          // texture coordinate indices
     vertexList = NULL;  // vertices
}    




// CMD2Model::SetupSkin()
// access: private
// desc: sets up the model skin/texture for OpenGL
void SetupSkin(texture_t thisTexture)
{
     // set the proper parameters for an MD2 texture
     glGenTextures(1, &thisTexture.texID);
     glBindTexture(GL_TEXTURE_2D, thisTexture.texID);
     glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
     glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
     glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
     glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
     
     switch (thisTexture.textureType)
     {
     case BMP:
          gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, thisTexture.width, thisTexture.height,
               GL_RGB, GL_UNSIGNED_BYTE, thisTexture.data);
          break;
     case TGA:
          break;
     default:
          break;
     }
}

// CMD2Model::Load()
// access: public
// desc: loads model and skin

void initList(){
     modellist=glGenLists(1);
     glNewList(modellist,GL_COMPILE);
     RenderFrame(0);
     glEndList();

}

int Load(){
    modelfile_t minf;
    sprintf(minf.info,"%s",name);
    Load(&minf);
    initList();
    return true;
}
    
int Load(char *minfoname){
    modelfile_t minf;
    sprintf(minf.info,"%s",minfoname);
    Load(&minf);
    initList();
    return true;
}

int Load(modelfile_t *modelinfo)
{
     FILE *md2conf;
     int temp;
     md2conf=fopen(modelinfo->info,"rt");
     if(md2conf==NULL) return false;
     fscanf(md2conf,"MODELFILE: %s\n",&modelinfo->name);
     fscanf(md2conf,"MODELTEX: %s\n",&modelinfo->tex);
     char tmp[50];
     fscanf(md2conf,"OVERBLEND BLACK?: %i\n",&temp);
     if(temp==1)alpha=true;
     else alpha=false;
     fclose(md2conf);
     init();
     FILE *filePtr;                          // file pointer
     int fileLen;                            // length of model file
     char *buffer;                           // file buffer
     
     modelHeader_t *modelHeader;             // model header
     
     stIndex_t *stPtr;                       // texture data
     frame_t *frame;                              // frame data
     Vertex *vertexListPtr;                // index variable
     mesh_t *bufIndexPtr;          // index variables
     int i, j;                               // index variables
     
     // open the model file
     filePtr = fopen(modelinfo->name, "rb");
     if (filePtr == NULL)
          return FALSE;
     
     // find length of file
     fseek(filePtr, 0, SEEK_END);
     fileLen = ftell(filePtr);
     fseek(filePtr, 0, SEEK_SET);
     
     // read entire file into buffer
     buffer = new char [fileLen+1];
     fread(buffer, sizeof(char), fileLen, filePtr);
     
     // extract model file header from buffer
     modelHeader = (modelHeader_t*)buffer;
     
     vertexList = new Vertex [modelHeader->numXYZ * modelHeader->numFrames];
     
     numVertices = modelHeader->numXYZ;
     numFrames = modelHeader->numFrames;
     frameSize = modelHeader->framesize;
     
     for (j = 0; j < numFrames; j++)
     {
          frame = (frame_t*)&buffer[modelHeader->offsetFrames + frameSize * j];
          
          vertexListPtr = (Vertex*)&vertexList[numVertices * j];
          for (i = 0; i < numVertices; i++)
          {
               vertexListPtr[i].x = frame->scale[0] * frame->fp[i].v[0] + frame->translate[0];
               vertexListPtr[i].y = frame->scale[1] * frame->fp[i].v[1] + frame->translate[1];
               vertexListPtr[i].z = frame->scale[2] * frame->fp[i].v[2] + frame->translate[2];
          }
     }
     
     modelTex = LoadTexture(modelinfo->tex);

     SetupSkin(modelTex);

     
     numST = modelHeader->numST;
     st = new texCoord_t [numST];
     
     stPtr = (stIndex_t*)&buffer[modelHeader->offsetST];
     for (i = 0; i < numST; i++)
     {
          st[i].s = (float)stPtr[i].s / (float)modelTex.width;
          st[i].t = (float)stPtr[i].t / (float)modelTex.height;
     }
     
     numTriangles = modelHeader->numTris;
     triIndex = new mesh_t [numTriangles];
     
     // point to triangle indexes in buffer
     bufIndexPtr = (mesh_t*)&buffer[modelHeader->offsetTris];
     
     // create a mesh (triangle) list
     for (j = 0; j < numFrames; j++)         
     {
          // for all triangles in each frame
          for(i = 0; i < numTriangles; i++)
          {
               triIndex[i].meshIndex[0] = bufIndexPtr[i].meshIndex[0];
               triIndex[i].meshIndex[1] = bufIndexPtr[i].meshIndex[1];
               triIndex[i].meshIndex[2] = bufIndexPtr[i].meshIndex[2];
               triIndex[i].stIndex[0] = bufIndexPtr[i].stIndex[0];
               triIndex[i].stIndex[1] = bufIndexPtr[i].stIndex[1];
               triIndex[i].stIndex[2] = bufIndexPtr[i].stIndex[2];
          }
     }
     
     // close file and free memory
     fclose(filePtr);
     free(buffer);
     
     currentFrame = 0;
     nextFrame = 1;
     interpol = 0.0;
     
     return TRUE;
}

// CMD2Model::LoadModel()
// access: public
// desc: loads model from file
int LoadModel(char *modelFile)
{
     FILE *filePtr;                          // file pointer
     int fileLen;                            // length of model file
     char *buffer;                           // file buffer
     
     modelHeader_t *modelHeader;             // model header
     
     stIndex_t *stPtr;                       // texture data
     frame_t *frame;                              // frame data
     Vertex *vertexListPtr;                // index variable
     mesh_t *triIndex, *bufIndexPtr;         // index variables
     int i, j;                               // index variables
     
     // open the model file
     filePtr = fopen(modelFile, "rb");
     if (filePtr == NULL)
          return FALSE;
     
     // find length of file
     fseek(filePtr, 0, SEEK_END);
     fileLen = ftell(filePtr);
     fseek(filePtr, 0, SEEK_SET);
     
     // read entire file into buffer
     buffer = new char [fileLen+1];
     fread(buffer, sizeof(char), fileLen, filePtr);
     
     // extract model file header from buffer
     modelHeader = (modelHeader_t*)buffer;
     
     // allocate vertex list
     vertexList = new Vertex [modelHeader->numXYZ * modelHeader->numFrames];
     
     numVertices = modelHeader->numXYZ;
     numFrames = modelHeader->numFrames;
     frameSize = modelHeader->framesize;
     
     for (j = 0; j < numFrames; j++)
     {
          frame = (frame_t*)&buffer[modelHeader->offsetFrames + frameSize * j];
          
          vertexListPtr = (Vertex*)&vertexList[numVertices * j];
          for (i = 0; i < numVertices; i++)
          {
               vertexListPtr[i].x = frame->scale[0] * frame->fp[i].v[0] + frame->translate[0];
               vertexListPtr[i].y = frame->scale[1] * frame->fp[i].v[1] + frame->translate[1];
               vertexListPtr[i].z = frame->scale[2] * frame->fp[i].v[2] + frame->translate[2];
          }
     }
     
     numST = modelHeader->numST;
     
     st = new texCoord_t [numST];
     
     stPtr = (stIndex_t*)&buffer[modelHeader->offsetST];
     for (i = 0; i < numST; i++)
     {
          st[i].s = 0.0;
          st[i].t = 0.0;
     }
     
     numTriangles = modelHeader->numTris;
     triIndex = new mesh_t [numTriangles];
     
     // point to triangle indexes in buffer
     bufIndexPtr = (mesh_t*)&buffer[modelHeader->offsetTris];
     
     // create a mesh (triangle) list
     for (j = 0; j < numFrames; j++)         
     {
          // for all triangles in each frame
          for(i = 0; i < numTriangles; i++)
          {
               triIndex[i].meshIndex[0] = bufIndexPtr[i].meshIndex[0];
               triIndex[i].meshIndex[1] = bufIndexPtr[i].meshIndex[1];
               triIndex[i].meshIndex[2] = bufIndexPtr[i].meshIndex[2];
               triIndex[i].stIndex[0] = bufIndexPtr[i].stIndex[0];
               triIndex[i].stIndex[1] = bufIndexPtr[i].stIndex[1];
               triIndex[i].stIndex[2] = bufIndexPtr[i].stIndex[2];
          }
     }
     
     // close file and free memory
     fclose(filePtr);
     

     currentFrame = 0;
     nextFrame = 1;
     interpol = 0.0;
     
     
     
     return 0;
}

// CMD2Model::LoadSkin()
// access: public
// desc: loads a skin for the model
int LoadSkin(char *skinFile)
{
     int i;
     
     modelTex = LoadTexture(skinFile);
     
     SetupSkin(modelTex);

     
     for (i = 0; i < numST; i++)
     {
          st[i].s /= (float)modelTex.width;
          st[i].t /= (float)modelTex.height;
     }
     
     return 0;
}

// CMD2Model::SetTexture()
// access: public
// desc: sets a new texture object 
int SetTexture(texture_t texture)
{
     int i;
     

     modelTex = texture;

     
     SetupSkin(modelTex);
     
     for (i = 0; i < numST; i++)
     {
          st[i].s /= (float)modelTex.width;
          st[i].t /= (float)modelTex.height;
     }
     
     return 0;
}

// CMD2Model::Animate()
// access: public
// desc: animates the model between the keyframes startFrame and endFrame
int Animate(int startFrame, int endFrame, float percent)
{
     Vertex *vList;              // current frame vertices
     Vertex *nextVList;          // next frame vertices
     int i;                                  // index counter
     float x1, y1, z1;                  // current frame point values
     float x2, y2, z2;                  // next frame point values
     
     Vertex vertex[3];
     
     if ((startFrame > currentFrame))
          currentFrame = startFrame;
     
     if ((startFrame < 0) || (endFrame < 0))
          return -1;
     
     if ((startFrame >= numFrames) || (endFrame >= numFrames))
          return -1;
     
     if (interpol >= 1.0)
     {
          interpol = 0.0f;
          currentFrame++;
          if (currentFrame >= endFrame)
               currentFrame = startFrame;
          
          nextFrame = currentFrame + 1;
          
          if (nextFrame >= endFrame)
               nextFrame = startFrame;
          
     }
     
     vList = &vertexList[numVertices*currentFrame];
     nextVList = &vertexList[numVertices*nextFrame];
    
     
     glBindTex(modelTex);
     glBegin(GL_TRIANGLES);
     for (i = 0; i < numTriangles; i++)
     {
          // get first points of each frame
          x1 = vList[triIndex[i].meshIndex[0]].x;
          y1 = vList[triIndex[i].meshIndex[0]].y;
          z1 = vList[triIndex[i].meshIndex[0]].z;
          x2 = nextVList[triIndex[i].meshIndex[0]].x;
          y2 = nextVList[triIndex[i].meshIndex[0]].y;
          z2 = nextVList[triIndex[i].meshIndex[0]].z;
          
          // store first interpolated vertex of triangle
          vertex[0].x = x1 + interpol * (x2 - x1);
          vertex[0].y = y1 + interpol * (y2 - y1);
          vertex[0].z = z1 + interpol * (z2 - z1);
          
          // get second points of each frame
          x1 = vList[triIndex[i].meshIndex[2]].x;
          y1 = vList[triIndex[i].meshIndex[2]].y;
          z1 = vList[triIndex[i].meshIndex[2]].z;
          x2 = nextVList[triIndex[i].meshIndex[2]].x;
          y2 = nextVList[triIndex[i].meshIndex[2]].y;
          z2 = nextVList[triIndex[i].meshIndex[2]].z;
          
          // store second interpolated vertex of triangle
          vertex[2].x = x1 + interpol * (x2 - x1);
          vertex[2].y = y1 + interpol * (y2 - y1);
          vertex[2].z = z1 + interpol * (z2 - z1);
          
          // get third points of each frame
          x1 = vList[triIndex[i].meshIndex[1]].x;
          y1 = vList[triIndex[i].meshIndex[1]].y;
          z1 = vList[triIndex[i].meshIndex[1]].z;
          x2 = nextVList[triIndex[i].meshIndex[1]].x;
          y2 = nextVList[triIndex[i].meshIndex[1]].y;
          z2 = nextVList[triIndex[i].meshIndex[1]].z;
          
          // store third interpolated vertex of triangle
          vertex[1].x = x1 + interpol * (x2 - x1);
          vertex[1].y = y1 + interpol * (y2 - y1);
          vertex[1].z = z1 + interpol * (z2 - z1);
          
          // calculate the normal of the triangle
          CalculateNormal(vertex[0], vertex[2], vertex[1]);
          
          // render properly textured triangle
          glTexCoord2f(st[triIndex[i].stIndex[0]].s,
               st[triIndex[i].stIndex[0]].t);
          glVertex3f(vertex[0].x,vertex[0].y,vertex[0].z);
          
          glTexCoord2f(st[triIndex[i].stIndex[2]].s ,
               st[triIndex[i].stIndex[2]].t);
          glVertex3f(vertex[2].x,vertex[2].y,vertex[2].z);
          
          glTexCoord2f(st[triIndex[i].stIndex[1]].s,
               st[triIndex[i].stIndex[1]].t);
          glVertex3f(vertex[1].x,vertex[1].y,vertex[1].z);
     }
     glEnd();
     
     interpol += percent;  // increase percentage of interpolation between frames
     
     return 0;
}

// RenderFrame()
// desc: renders a single key frame
void RenderFrame(){
    glCallList(modellist);
}

int RenderFrame(int keyFrame)
{
     Vertex *vList;
     int i;
     
     // create a pointer to the frame we want to show
     vList = &vertexList[numVertices * keyFrame];
     
     
		// set the texture
	 glBindTex(modelTex);
     
     // display the textured model with proper lighting normals
     glBegin(GL_TRIANGLES);
     for(i = 0; i < numTriangles; i++)
     {
          CalculateNormal(vList[triIndex[i].meshIndex[0]],
               vList[triIndex[i].meshIndex[2]],
               vList[triIndex[i].meshIndex[1]]);


			glTexCoord2f(st[triIndex[i].stIndex[0]].s,
				st[triIndex[i].stIndex[0]].t);

          glVertex3f(vList[triIndex[i].meshIndex[0]].x,vList[triIndex[i].meshIndex[0]].y,vList[triIndex[i].meshIndex[0]].z);
          

			glTexCoord2f(st[triIndex[i].stIndex[2]].s ,
				st[triIndex[i].stIndex[2]].t);

          glVertex3f(vList[triIndex[i].meshIndex[2]].x,vList[triIndex[i].meshIndex[2]].y,vList[triIndex[i].meshIndex[2]].z);
          

			glTexCoord2f(st[triIndex[i].stIndex[1]].s,
				st[triIndex[i].stIndex[1]].t);

          glVertex3f(vList[triIndex[i].meshIndex[1]].x,vList[triIndex[i].meshIndex[1]].y,vList[triIndex[i].meshIndex[1]].z);
     }
     glEnd();
     
     return 0;
}

int RenderShadowedFrame(Vertex lightpos,int keyFrame=0)
{
     Vertex *vList;
     Vertex v;
     int i;
     int LENGTH=500;
     
     // create a pointer to the frame we want to show
     vList = &vertexList[numVertices * keyFrame];
     
     
		// set the texture
	 //glBindTex(modelTex);
	 glDisable(GL_TEXTURE_2D);
     
     // display the textured model with proper lighting normals
     glBegin(GL_QUADS);
     for(i = 0; i < numTriangles; i++)
     {
          //1. Viereck
          glVertex3f(vList[triIndex[i].meshIndex[0]].x,vList[triIndex[i].meshIndex[0]].y,vList[triIndex[i].meshIndex[0]].z);
          
          v=newV(vList[triIndex[i].meshIndex[0]].x,vList[triIndex[i].meshIndex[0]].y,vList[triIndex[i].meshIndex[0]].z);
          v=sub(v,lightpos);
          normalizeVec(&v);
          v=multScalar(v,LENGTH);
          glVertexV(v);        
          
          v=newV(vList[triIndex[i].meshIndex[2]].x,vList[triIndex[i].meshIndex[2]].y,vList[triIndex[i].meshIndex[2]].z);
          v=sub(v,lightpos);
          normalizeVec(&v);
          v=multScalar(v,LENGTH);
          glVertexV(v);          
          glVertex3f(vList[triIndex[i].meshIndex[2]].x,vList[triIndex[i].meshIndex[2]].y,vList[triIndex[i].meshIndex[2]].z);
          
          
          //2.Viereck
          glVertex3f(vList[triIndex[i].meshIndex[1]].x,vList[triIndex[i].meshIndex[1]].y,vList[triIndex[i].meshIndex[1]].z);

          v=newV(vList[triIndex[i].meshIndex[1]].x,vList[triIndex[i].meshIndex[1]].y,vList[triIndex[i].meshIndex[1]].z);
          v=sub(v,lightpos);
          normalizeVec(&v);
          v=multScalar(v,LENGTH);
          glVertexV(v);        
          
          v=newV(vList[triIndex[i].meshIndex[2]].x,vList[triIndex[i].meshIndex[2]].y,vList[triIndex[i].meshIndex[2]].z);
          v=sub(v,lightpos);
          normalizeVec(&v);
          v=multScalar(v,LENGTH);
          glVertexV(v);          
          
          glVertex3f(vList[triIndex[i].meshIndex[2]].x,vList[triIndex[i].meshIndex[2]].y,vList[triIndex[i].meshIndex[2]].z);
          
          //3. Viereck
          glVertex3f(vList[triIndex[i].meshIndex[0]].x,vList[triIndex[i].meshIndex[0]].y,vList[triIndex[i].meshIndex[0]].z);
          
          v=newV(vList[triIndex[i].meshIndex[0]].x,vList[triIndex[i].meshIndex[0]].y,vList[triIndex[i].meshIndex[0]].z);
          v=sub(v,lightpos);
          normalizeVec(&v);
          v=multScalar(v,LENGTH);
          glVertexV(v);        
          
          v=newV(vList[triIndex[i].meshIndex[1]].x,vList[triIndex[i].meshIndex[1]].y,vList[triIndex[i].meshIndex[1]].z);
          v=sub(v,lightpos);
          normalizeVec(&v);
          v=multScalar(v,LENGTH);
          glVertexV(v);          
          glVertex3f(vList[triIndex[i].meshIndex[1]].x,vList[triIndex[i].meshIndex[1]].y,vList[triIndex[i].meshIndex[1]].z);
     }
     glEnd();
     
     glEnable(GL_TEXTURE_2D);
     
     return 0;
}


// Unload()
// desc: unloads model data from memory
int Unload()
{
     if (triIndex != NULL)
          free(triIndex);
     if (vertexList != NULL)
          free(vertexList);
     if (st != NULL)
          free(st);
     
     return 0;
}


};

//**************************************************************
//ModelInstance
//***************************************************************
class ModelInstance{
      public:
           Vertex pos,rot;
           float scale,anim;
           int parentNR;
     
      
      void SetModelPos(){
           
           glPushMatrix();
           glTranslatev(pos);
           glRotatev(rot);
           glRotatef(-90,1,0,0);
           glScalef(scale,scale,scale);
           
      }
      
      void NextFrame(){
           anim+=0.25;
      }
      
      
};

//**********************************************************
//Renders a billboard with texturecoordinates
//**********************************************************

void RenderBillboard(float dx,float dz,int TexNr=0){
     glBegin(GL_QUADS);
            glTexCoordARB(0,0,TexNr);glVertex3f(-dz,-0.5,-dx);
            glTexCoordARB(1,0,TexNr);glVertex3f(dz,-0.5,dx);
            glTexCoordARB(1,1,TexNr);glVertex3f(dz,0.5,dx);
            glTexCoordARB(0,1,TexNr);glVertex3f(dz,-0.5,dx);
     glEnd();                    
}

#endif
